#!/usr/bin/python -tt
#
# Fullly virtualized guest support
#
# Copyright 2006  Red Hat, Inc.
# Jeremy Katz <katzj@redhat.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os,stat,time,sys
import string

import libvirt

import XenGuest
import util

if os.uname()[4] in ("x86_64"):
    qemu = "/usr/lib64/xen/bin/qemu-dm"
else:
    qemu = "/usr/lib/xen/bin/qemu-dm"

class FullVirtGuest(XenGuest.XenGuest):
    OS_TYPES = { "linux": { "label": "Linux", \
                            "acpi": True, \
                            "apic": True, \
                            "continue": False, \
                            "timer_mode": None, \
                            "input": [ "mouse", "ps2"],
                            "variants": { "rhel2.1": { "label": "Red Hat Enterprise Linux 2.1", "distro": "rhel", "input": ["mouse","ps2"] }, \
                                          "rhel3": { "label": "Red Hat Enterprise Linux 3", "distro": "rhel" }, \
                                          "rhel3_64": { "label": "Red Hat Enterprise Linux 3 64-bit", "distro": "rhel" }, \
                                          "rhel4": { "label": "Red Hat Enterprise Linux 4", "distro": "rhel", "input": ["mouse","ps2"] }, \
                                          "rhel4_64": { "label": "Red Hat Enterprise Linux 4 64-bit", "distro": "rhel", "timer_mode": "2", "input": ["mouse","ps2"] }, \
                                          "rhel5": { "label": "Red Hat Enterprise Linux 5", "distro": "rhel" }, \
                                          "rhel5_64": { "label": "Red Hat Enterprise Linux 5 64-bit", "distro": "rhel", "timer_mode": "2" }, \
                                          "el4": { "label": "Oracle Enterprise Linux 4", "distro": "rhel", "input": ["mouse","ps2"] }, \
                                          "el4_64": { "label": "Oracle Enterprise Linux 4 64-bit", "distro": "rhel", "timer_mode": "2", "input": ["mouse","ps2"] }, \
                                          "el5": { "label": "Oracle Enterprise Linux 5", "distro": "rhel" }, \
                                          "el5_64": { "label": "Oracle Enterprise Linux 5 64-bit", "distro": "rhel", "timer_mode": "2" }, \
                                          "centos5": { "label": "Cent OS 5", "distro": "centos" }, \
                                          "centos5_64": { "label": "Cent OS 5 64-bit", "distro": "centos", "timer_mode": "2" }, \
                                          "fedora5": { "label": "Fedora Core 5", "distro": "fedora" }, \
                                          "fedora5_64": { "label": "Fedora Core 5 64-bit", "distro": "fedora", "timer_mode": "2"  }, \
                                          "fedora6": { "label": "Fedora Core 6", "distro": "fedora" }, \
                                          "fedora6_64": { "label": "Fedora Core 6 64-bit", "distro": "fedora", "timer_mode": "2"  }, \
                                          "fedora7": { "label": "Fedora 7", "distro": "fedora" }, \
                                          "fedora7_64": { "label": "Fedora 7 64-bit", "distro": "fedora", "timer_mode": "2"  }, \
                                          "sles10": { "label": "Suse Linux Enterprise Server", "distro": "suse" }, \
                                          "sles10_64": { "label": "Suse Linux Enterprise Server 64-bit", "distro": "suse", "timer_mode": "2" }, \
                                          "generic24": { "label": "Generic 2.4.x kernel" }, \
                                          "generic26": { "label": "Generic 2.6.x kernel" }, \
                                          "generic26_64": { "label": "Generic 2.6.x kernel 64-bit", "timer_mode": "2" }, \
                                          }, \
                            }, \
                 "windows": { "label": "Windows", \
                              "acpi": True, \
                              "apic": True, \
                              "continue": True, \
                              "timer_mode": "1", \
                              "input": [ "tablet", "usb"],
                              "variants": { "winxp": { "label": "Microsoft Windows XP", \
                                                       "acpi": False, \
                                                       "apic": False }, \
                                            "win2k": { "label": "Microsoft Windows 2000", \
                                                       "acpi": False, \
                                                       "apic": False }, \
                                            "win2k3": { "label": "Microsoft Windows 2003" }, \
                                            "vista": { "label": "Microsoft Windows Vista" }, \
                                            }, \
                              }, \
                 "unix": { "label": "UNIX", \
                           "acpi": True,
                           "apic": True,
                           "continue": False, \
                           "input": [ "mouse", "ps2"],
                           "variants": { "solaris9": { "label": "Sun Solaris 9" }, \
                                         "solaris10": { "label": "Sun Solaris 10" }, \
                                         "freebsd6": { "label": "Free BSD 6.x" }, \
                                         "openbsd4": { "label": "Open BSD 4.x" }, \
                                         }, \
                           }, \
                 "other": { "label": "Other", \
                            "acpi": True,
                            "apic": True,
                            "continue": False,
                            "input": [ "mouse", "ps2"],
                            "variants": { "msdos": { "label": "MS-DOS", \
                                                     "acpi": False, \
                                                     "apic": False }, \
                                          "netware4": { "label": "Novell Netware 4" }, \
                                          "netware5": { "label": "Novell Netware 5" }, \
                                          "netware6": { "label": "Novell Netware 6" }, \
                                          "generic": { "label": "Generic" }, \
                                          }, \
                            } \
                 }

    def list_os_types():
        return FullVirtGuest.OS_TYPES.keys()
    list_os_types = staticmethod(list_os_types)

    def list_os_variants(type):
        return FullVirtGuest.OS_TYPES[type]["variants"].keys()
    list_os_variants = staticmethod(list_os_variants)

    def get_os_type_label(type):
        return FullVirtGuest.OS_TYPES[type]["label"]
    get_os_type_label = staticmethod(get_os_type_label)

    def get_os_variant_label(type, variant):
        return FullVirtGuest.OS_TYPES[type]["variants"][variant]["label"]
    get_os_variant_label = staticmethod(get_os_variant_label)

    def __init__(self, hypervisorURI=None):
        XenGuest.XenGuest.__init__(self, hypervisorURI=hypervisorURI)
        self._cdrom = None
        self.havecdrom = False
        self.keepcdimage = False
        self.disknode = "hd"
        self.features = { "acpi": None, "pae": util.is_pae_capable(), "apic": None, "timer_mode": None }
        self._os_type = None
        self._os_variant = None
        self._vif_type = "ioemu"

    def get_cdrom(self):
        return self._cdrom
    def set_cdrom(self, val):
        val = os.path.abspath(val)
        if not os.path.exists(val):
            raise ValueError, "CD device must exist!"
        self._cdrom = val
    cdrom = property(get_cdrom, set_cdrom)

    def _get_disk_xml(self):
        # ugh, this is disgusting, but the HVM disk stuff isn't nice :/
        xml = XenGuest.XenGuest._get_disk_xml(self)
        if self.cdrom:
            disk = XenGuest.XenDisk(self.cdrom, readOnly = True, device=XenGuest.XenDisk.DEVICE_CDROM)
            # XXX no need to hardcode hdc in newer xen
            xml += disk.get_xml_config("hdc")
        return xml

    def get_os_type(self):
        return self._os_type
    def set_os_type(self, val):
        if FullVirtGuest.OS_TYPES.has_key(val):
            self._os_type = val
        else:
            raise RuntimeError, "OS type %s does not exist in our dictionary" % val
    os_type = property(get_os_type, set_os_type)

    def get_os_variant(self):
        return self._os_variant
    def set_os_variant(self, val):
        if not self._os_type:
            raise ValueError, _("An OS type must be specified before a variant.")
        if FullVirtGuest.OS_TYPES[self._os_type]["variants"].has_key(val):
            self._os_variant = val
        else:
            raise RuntimeError, "OS variant %s does not exist in our dictionary for OS type %s" % (val, os_type)
    os_variant = property(get_os_variant, set_os_variant)

    def set_os_type_parameters(self, os_type, os_variant):
        # explicitly disabling apic and acpi will override OS_TYPES values
        if self.features["acpi"] is None and os_type is not None:
            if os_variant is not None and FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant].has_key("acpi"):
                self.features["acpi"] = FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant]["acpi"]
            else:
                self.features["acpi"] = FullVirtGuest.OS_TYPES[os_type]["acpi"]
        elif self.features["acpi"] is None:
            # default for acpi should be "true", but we can't set that until we get here
            self.features["acpi"] = True

        if self.features["apic"] is None and os_type is not None:
            if os_variant is not None and FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant].has_key("apic"):
                self.features["apic"] = FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant]["apic"]
            else:
                self.features["apic"] = FullVirtGuest.OS_TYPES[os_type]["apic"]
        elif self.features["apic"] is None:
            # default for apic should be "true", but we can't set that until we get here
            self.features["apic"] = True

        if self.features["timer_mode"] is None and os_type is not None:
            if os_variant is not None and FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant].has_key("timer_mode"):
                self.features["timer_mode"] = FullVirtGuest.OS_TYPES[os_type]["variants"][os_variant]["timer_mode"]
            else:
                self.features["timer_mode"] = FullVirtGuest.OS_TYPES[os_type]["timer_mode"]
        elif self.features["timer_mode"] is None:
            # default for timer_mode should be "0", but we can't set that until we get here
            self.features["timer_mode"] = "0"

    def get_os_distro(self):
        if self.os_type is not None and self.os_variant is not None and "distro" in FullVirtGuest.OS_TYPES[self.os_type]["variants"][self.os_variant]:
            return FullVirtGuest.OS_TYPES[self.os_type]["variants"][self.os_variant]["distro"]
        return None
    os_distro = property(get_os_distro)
 
    def get_vif_type(self):
        return self._vif_type
    def set_vif_type(self, val):
        if val in ("ioemu", "netfront"):
            self._vif_type = val
        else:
            raise ValueError, "Vif type must be 'ioemu' or 'netfront'"

    vif_type = property(get_vif_type, set_vif_type)
 
    def _get_disk_xen(self):
        # We are again hacking this code.  We only do it for the cfg because
        # the XML side is hacking the boot cd.
        # Keep CD-ROM image during installation
        xen = XenGuest.XenGuest._get_disk_xen(self)
        if self.keepcdimage:
            xen = xen[:-1] + '\'file:%s,hdc:cdrom,r\']' % self.cdrom
        elif self.havecdrom:
            xen = xen[:-1] + '\',hdc:cdrom,r\', ]'

        return xen

    def _get_features_xml(self):
        ret = ""
        for (k, v) in self.features.items():
            if v:
                if v != False:
                    ret += "<%s>%s</%s>" %(k,v,k)
                else:
                    ret += "<%s/>" %(k,)
        return ret

    def _get_features_xen(self):
        ret = ""
        for (k, v) in self.features.items():
            if v:
                if v == True:
                    ret += "%s=1\n" %(k,)
                elif v == False:
                    ret += "%s=0\n" %(k,)
                else:
                    ret += "%s=%s\n" %(k,v)
        return ret

    def _get_network_xen(self):
        # Sick copying this from XenGuest.py, but ultimately all
        # XM config stuff will be moved to xend/libvirt and so this
        # will go away.
        """Get the network config in the xend python format"""
        if len(self.nics) == 0: return ""
        # all vif types are either ioemu or netfront for now
        ret = "vif = [ "
        for n in self.nics:
            ret += "'type=%(viftype)s, mac=%(mac)s, bridge=%(bridge)s', " % { "bridge": n.bridge, "mac": n.macaddr, "viftype": self._vif_type }
        ret += "]"
        return ret

    def get_input_device(self):
        input = ("mouse", "ps2")
        if self.os_type is not None:
            if self.os_variant is not None and FullVirtGuest.OS_TYPES[self.os_type]["variants"][self.os_variant].has_key("input"):
                input = FullVirtGuest.OS_TYPES[self.os_type]["variants"][self.os_variant]["input"]
            elif FullVirtGuest.OS_TYPES[self.os_type].has_key("input"):
                input = FullVirtGuest.OS_TYPES[self.os_type]["input"]
        return (input[0], input[1])

    def _get_config_xml(self, install = True):
        # FIXME: hard-codes that we're booting from CD as hdd
        if install:
            action = "destroy"
            bootdev = "cdrom"
        else:
            action = "restart"
            bootdev = "hd"

        return """<domain type='xen'>
  <name>%(name)s</name>
  <os>
    <type>hvm</type>
    <loader>/usr/lib/xen/boot/hvmloader</loader>
    <boot dev='%(bootdev)s'/>
  </os>
  <features>
    %(features)s
  </features>
  <memory>%(ramkb)s</memory>
  <vcpu>%(vcpus)d</vcpu>
  <uuid>%(uuid)s</uuid>
  <on_reboot>%(action)s</on_reboot>
  <on_crash>%(action)s</on_crash>
  <on_poweroff>destroy</on_poweroff>
  <devices>
    <emulator>%(qemu)s</emulator>
%(disks)s
    %(networks)s
    %(graphics)s
  </devices>
</domain>
""" % { "qemu": qemu, "name": self.name, "vcpus": self.vcpus, "uuid": self.uuid, "ramkb": self.memory * 1024, "disks": self._get_disk_xml(), "networks": self._get_network_xml(), "graphics": self._get_graphics_xml(), "features": self._get_features_xml(), "action": action, "bootdev": bootdev }

    def _get_config_xen(self):
        return """# Automatically generated xen config file
name = "%(name)s"
builder = "hvm"
memory = "%(ram)s"
%(disks)s
%(networks)s
uuid = "%(uuid)s"
device_model = "%(qemu)s"
kernel = "/usr/lib/xen/boot/hvmloader"
%(graphics)s
%(features)s
vcpus=%(vcpus)s
serial = "pty" # enable serial console
on_reboot   = 'restart'
on_crash    = 'restart'
""" % { "name": self.name, "ram": self.memory, "disks": self._get_disk_xen(), "networks": self._get_network_xen(), "uuid": self.uuid, "qemu": qemu, "graphics": self._get_graphics_xen(), "vcpus": self.vcpus, "features": self._get_features_xen() }

    def validate_parms(self):
        if not self.cdrom:
            raise RuntimeError, "A CD must be specified to boot from"
        self.set_os_type_parameters(self.os_type, self.os_variant)
        XenGuest.XenGuest.validate_parms(self)
